import unittest
import errno


from pyverbs.qp import QP, QPAttr, QPInitAttr, QPCap
from pyverbs.pyverbs_error import PyverbsRDMAError
from tests.base import BaseResources, RDMATestCase
from pyverbs.providers.mlx5.mlx5dv import Mlx5QP
from tests.utils import requires_root_on_eth
import pyverbs.enums as e
from pyverbs.cq import CQ


class LagRawQP(BaseResources):
    def __init__(self, dev_name):
        super().__init__(dev_name, None, None)
        self.cq = self.create_cq()
        self.qp = self.create_qp()

    def create_cq(self):
        return CQ(self.ctx, 100)

    @requires_root_on_eth()
    def create_qp(self):
        qia = QPInitAttr(e.IBV_QPT_RAW_PACKET, rcq=self.cq, scq=self.cq,
                         cap=QPCap())
        try:
            qp = QP(self.pd, qia)
        except PyverbsRDMAError as ex:
            if ex.error_code == errno.EOPNOTSUPP:
                raise unittest.SkipTest("Create Raw Packet QP is not supported")
            raise ex
        qp.to_init(QPAttr())
        return qp


class LagPortTestCase(RDMATestCase):
    def modify_lag(self, resources):
        try:
            port_num, active_port_num = Mlx5QP.query_lag_port(resources.qp)
            # if port_num is 1 - modify to 2, else modify to 1
            new_port_num = (2 - port_num) + 1
            Mlx5QP.modify_lag_port(resources.qp, new_port_num)
            port_num, active_port_num = Mlx5QP.query_lag_port(resources.qp)
            self.assertEqual(port_num, new_port_num, 'Port num is not as expected')
        except PyverbsRDMAError as ex:
            if ex.error_code == errno.EOPNOTSUPP:
                raise unittest.SkipTest('Set LAG affinity is not supported on this device')
            raise ex

    def test_raw_modify_lag_port(self):
        qp = LagRawQP(self.dev_name)
        self.modify_lag(qp)



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019 Mellanox Technologies, Inc. All rights reserved. See COPYING file
"""
Test module for pyverbs' cq module.
"""
import random

from pyverbs.pyverbs_error import PyverbsError, PyverbsRDMAError
from pyverbs.cq import CompChannel, CQ, CqInitAttrEx, CQEX
from tests.base import PyverbsAPITestCase
import pyverbs.enums as e
import unittest
import errno


class CQTest(PyverbsAPITestCase):
    """
    Test various functionalities of the CQ class.
    """
    def test_create_cq(self):
        """
        Test ibv_create_cq()
        """
        for ctx, attr, attr_ex in self.devices:
            for i in range(10):
                cqes = get_num_cqes(attr)
                comp_vector = int(ctx.num_comp_vectors * random.random())
                if random.choice([True, False]):
                    with CompChannel(ctx) as cc:
                        with CQ(ctx, cqes, None, cc, comp_vector):
                            pass
                else:
                    with CQ(ctx, cqes, None, None, comp_vector):
                        pass

    def test_create_cq_bad_flow(self):
        """
        Test ibv_create_cq() with a wrong comp_vector / cqe number
        """
        for ctx, attr, attr_ex in self.devices:
            for i in range(10):
                cc = CompChannel(ctx)
                cqes = 100
                comp_vector = ctx.num_comp_vectors + int(100 *
                                                         random.random())
                has_cc = random.choice([True, False])
                if not has_cc:
                    cc = None
                try:
                    with CQ(ctx, cqes, None, cc, comp_vector):
                        pass
                except PyverbsError as ex:
                    assert 'Failed to create a CQ' in ex.args[0]
                    assert 'Invalid argument' in ex.args[0]
                else:
                    raise PyverbsError(
                        'Created a CQ with comp_vector={n} while device\'s num_comp_vectors={nc}'.
                        format(n=comp_vector, nc=ctx.num_comp_vectors))
                max_cqe = ctx.query_device().max_cqe
                cqes = random.randint(max_cqe + 1, max_cqe + 100)
                try:
                    with CQ(ctx, cqes, None, cc, 0):
                        pass
                except PyverbsError as ex:
                    assert 'Failed to create a CQ' in ex.args[0]
                    assert 'Invalid argument' in ex.args[0]
                else:
                    raise PyverbsError(
                        'Created a CQ with cqe={n} while device\'s max_cqe={nc}'.
                        format(n=cqes, nc=max_cqe))

    def test_destroy_cq(self):
        """
        Test ibv_destroy_cq()
        """
        for ctx, attr, attr_ex in self.devices:
            for i in range(10):
                cqes = get_num_cqes(attr)
                comp_vector = int(ctx.num_comp_vectors * random.random())
                if random.choice([True, False]):
                    with CompChannel(ctx) as cc:
                        cq = CQ(ctx, cqes, None, cc, comp_vector)
                else:
                    cq = CQ(ctx, cqes, None, None, comp_vector)
                cq.close()


class CCTest(PyverbsAPITestCase):
    """
    Test various functionalities of the Completion Channel class.
    """
    def test_create_comp_channel(self):
        """
        Test ibv_create_comp_channel()
        """
        for ctx, attr, attr_ex in self.devices:
            with CompChannel(ctx):
                pass

    def test_destroy_comp_channel(self):
        """
        Test ibv_destroy_comp_channel()
        """
        for ctx, attr, attr_ex in self.devices:
            cc = CompChannel(ctx)
            cc.close()


class CQEXTest(PyverbsAPITestCase):
    """
    Test various functionalities of the CQEX class.
    """
    def test_create_cq_ex(self):
        """
        Test ibv_create_cq_ex()
        """
        for ctx, attr, attr_ex in self.devices:
            cqe = get_num_cqes(attr)
            cq_init_attrs_ex = CqInitAttrEx(cqe=cqe, wc_flags=0, comp_mask=0, flags=0)
            wc_flags = get_cq_flags_with_caps()
            if attr_ex.raw_packet_caps & e.IBV_RAW_PACKET_CAP_CVLAN_STRIPPING == 0:
                wc_flags.remove(e.IBV_WC_EX_WITH_CVLAN)
            for f in wc_flags:
                cq_init_attrs_ex.wc_flags = f
                with CQEX(ctx, cq_init_attrs_ex):
                    pass
            # For the wc_flags that have no capability bit, we're not raising
            # an exception for EOPNOTSUPPORT
            wc_flags = get_cq_flags_with_no_caps()
            for f in wc_flags:
                cq_init_attrs_ex.wc_flags = f
                try:
                    with CQEX(ctx, cq_init_attrs_ex):
                        pass
                except PyverbsError as ex:
                    assert 'Failed to create extended CQ' in ex.args[0]
                    assert ' Errno: 95' in ex.args[0]
            cq_init_attrs_ex.wc_flags = 0
            cq_init_attrs_ex.comp_mask = e.IBV_CQ_INIT_ATTR_MASK_FLAGS
            attr_flags = list(e.ibv_create_cq_attr_flags)
            for f in attr_flags:
                cq_init_attrs_ex.flags = f
                try:
                    with CQEX(ctx, cq_init_attrs_ex):
                        pass
                except PyverbsError as ex:
                    assert 'Failed to create extended CQ' in ex.args[0]
                    assert ' Errno: 95' in ex.args[0]

    def test_create_cq_ex_bad_flow(self):
        """
        Test ibv_create_cq_ex() with wrong comp_vector / number of cqes
        """
        for ctx, attr, attr_ex in self.devices:
            for i in range(10):
                cq_attrs_ex = CqInitAttrEx(cqe=0, wc_flags=0, comp_mask=0, flags=0)
                max_cqe = attr.max_cqe
                cq_attrs_ex.cqe = max_cqe + 1 + int(100 * random.random())
                try:
                    CQEX(ctx, cq_attrs_ex)
                except PyverbsRDMAError as ex:
                    if ex.error_code == errno.EOPNOTSUPP:
                        raise unittest.SkipTest('Create extended CQ is not supported')
                    assert 'Failed to create extended CQ' in ex.args[0]
                    assert ' Errno: 22' in ex.args[0]
                else:
                    raise PyverbsError(
                        'Created a CQEX with {c} CQEs while device\'s max CQE={dc}'.
                        format(c=cq_attrs_ex.cqe, dc=max_cqe))
                comp_channel = random.randint(ctx.num_comp_vectors, 100)
                cq_attrs_ex.comp_vector = comp_channel
                cq_attrs_ex.cqe = get_num_cqes(attr)
                try:
                    CQEX(ctx, cq_attrs_ex)
                except PyverbsRDMAError as ex:
                    if ex.error_code == errno.EOPNOTSUPP:
                        raise unittest.SkipTest('Create extended CQ is not supported')
                    assert 'Failed to create extended CQ' in ex.args[0]
                    assert ' Errno: 22' in ex.args[0]
                else:
                    raise PyverbsError(
                        'Created a CQEX with comp_vector={c} while device\'s num_comp_vectors={dc}'.
                        format(c=comp_channel, dc=ctx.num_comp_vectors))

    def test_destroy_cq_ex(self):
        """
        Test ibv_destroy_cq() for extended CQs
        """
        for ctx, attr, attr_ex in self.devices:
            cqe = get_num_cqes(attr)
            cq_init_attrs_ex = CqInitAttrEx(cqe=cqe, wc_flags=0, comp_mask=0, flags=0)
            wc_flags = get_cq_flags_with_caps()
            if attr_ex.raw_packet_caps & e.IBV_RAW_PACKET_CAP_CVLAN_STRIPPING == 0:
                wc_flags.remove(e.IBV_WC_EX_WITH_CVLAN)
            for f in wc_flags:
                cq_init_attrs_ex.wc_flags = f
                with CQEX(ctx, cq_init_attrs_ex) as cq:
                    cq.close()
            # For the wc_flags that have no capability bit, we're not raising
            # an exception for EOPNOTSUPPORT
            wc_flags = get_cq_flags_with_no_caps()
            for f in wc_flags:
                cq_init_attrs_ex.wc_flags = f
                try:
                    with CQEX(ctx, cq_init_attrs_ex) as cq:
                        cq.close()
                except PyverbsError as ex:
                    assert 'Failed to create extended CQ' in ex.args[0]
                    assert ' Errno: 95' in ex.args[0]
            cq_init_attrs_ex.wc_flags = 0
            cq_init_attrs_ex.comp_mask = e.IBV_CQ_INIT_ATTR_MASK_FLAGS
            attr_flags = list(e.ibv_create_cq_attr_flags)
            for f in attr_flags:
                cq_init_attrs_ex.flags = f
                try:
                    with CQEX(ctx, cq_init_attrs_ex) as cq:
                        cq.close()
                except PyverbsError as ex:
                    assert 'Failed to create extended CQ' in ex.args[0]
                    assert ' Errno: 95' in ex.args[0]

def get_num_cqes(attr):
    max_cqe = attr.max_cqe
    return int((max_cqe + 1) * random.random())


def get_cq_flags_with_no_caps():
    wc_flags = list(e.ibv_create_cq_wc_flags)
    wc_flags.remove(e.IBV_WC_EX_WITH_CVLAN)
    return wc_flags


def get_cq_flags_with_caps():
    return [e.IBV_WC_EX_WITH_CVLAN]



from pyverbs.pyverbs_error import PyverbsRDMAError
from pyverbs.cq import CqInitAttrEx, CQEX
import pyverbs.enums as e
from pyverbs.mr import MR

from tests.base import RCResources, UDResources, XRCResources, RDMATestCase
import tests.utils as u
import unittest
import errno


def create_ex_cq(res):
    """
    Create an Extended CQ using res's context and assign it to res's cq member.
    IBV_WC_STANDARD_FLAGS is used for WC flags to avoid support differences
    between devices.
    :param res: An instance of TrafficResources
    """
    wc_flags = e.IBV_WC_STANDARD_FLAGS
    cia = CqInitAttrEx(cqe=2000, wc_flags=wc_flags)
    try:
        res.cq = CQEX(res.ctx, cia)
    except PyverbsRDMAError as ex:
        if ex.error_code == errno.EOPNOTSUPP:
            raise unittest.SkipTest('Create Extended CQ is not supported')
        raise ex

class CqExUD(UDResources):
    def create_cq(self):
        create_ex_cq(self)

    def create_mr(self):
        self.mr = MR(self.pd, self.msg_size + self.GRH_SIZE,
                     e.IBV_ACCESS_LOCAL_WRITE)


class CqExRC(RCResources):
    def create_cq(self):
        create_ex_cq(self)


class CqExXRC(XRCResources):
    def create_cq(self):
        create_ex_cq(self)


class CqExTestCase(RDMATestCase):
    """
    Run traffic over the existing UD, RC and XRC infrastructure, but use
    ibv_cq_ex instead of legacy ibv_cq
    """
    def setUp(self):
        super().setUp()
        self.iters = 100
        self.qp_dict = {'ud': CqExUD, 'rc': CqExRC, 'xrc': CqExXRC}

    def create_players(self, qp_type):
        client = self.qp_dict[qp_type](self.dev_name, self.ib_port,
                                       self.gid_index)
        server = self.qp_dict[qp_type](self.dev_name, self.ib_port,
                                       self.gid_index)
        client.pre_run(server.psns, server.qps_num)
        server.pre_run(client.psns, client.qps_num)
        return client, server

    def test_ud_traffic_cq_ex(self):
        client, server = self.create_players('ud')
        u.traffic(client, server, self.iters, self.gid_index, self.ib_port,
                  is_cq_ex=True)

    def test_rc_traffic_cq_ex(self):
        client, server = self.create_players('rc')
        u.traffic(client, server, self.iters, self.gid_index, self.ib_port,
                  is_cq_ex=True)

    def test_xrc_traffic_cq_ex(self):
        client, server = self.create_players('xrc')
        u.xrc_traffic(client, server, is_cq_ex=True)



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019 Mellanox Technologies, Inc. All rights reserved. See COPYING file
"""
Test module for Pyverbs' ParentDomain.
"""
from pyverbs.pd import ParentDomainInitAttr, ParentDomain, ParentDomainContext
from pyverbs.pyverbs_error import PyverbsRDMAError
from pyverbs.srq import SrqAttr, SrqInitAttr, SRQ
from pyverbs.cq import CqInitAttrEx, CQEX, CQ
from pyverbs.qp import QPInitAttr, QP
from tests.base import BaseResources
from tests.base import RDMATestCase
import pyverbs.mem_alloc as mem
import pyverbs.enums as e
import tests.utils as u
import unittest
import errno


class ParentDomainRes(BaseResources):
    def __init__(self, dev_name, ib_port=None, gid_index=None):
        super().__init__(dev_name=dev_name, ib_port=ib_port,
                         gid_index=gid_index)
        # Parent Domain will be created according to the test
        self.pd_ctx = None
        self.parent_domain = None


class ParentDomainTestCase(RDMATestCase):
    def setUp(self):
        super().setUp()
        self.pd_res = ParentDomainRes(self.dev_name)

    def _create_parent_domain_with_allocators(self, alloc_func, free_func):
        if alloc_func and free_func:
            self.pd_res.pd_ctx = ParentDomainContext(self.pd_res.pd, alloc_func,
                                                     free_func)
        pd_attr = ParentDomainInitAttr(pd=self.pd_res.pd,
                                       pd_context=self.pd_res.pd_ctx)
        try:
            self.pd_res.parent_domain = ParentDomain(self.pd_res.ctx,
                                                     attr=pd_attr)
        except PyverbsRDMAError as ex:
            if ex.error_code == errno.EOPNOTSUPP:
                raise unittest.SkipTest('Parent Domain is not supported on this device')
            raise ex

    def _create_rdma_objects(self):
        cq = CQ(self.pd_res.ctx, 100, None, None, 0)
        dev_attr = self.pd_res.ctx.query_device()
        qp_cap = u.random_qp_cap(dev_attr)
        qia = QPInitAttr(scq=cq, rcq=cq, cap=qp_cap)
        qia.qp_type = e.IBV_QPT_RC
        QP(self.pd_res.parent_domain, qia)
        srq_init_attr = SrqInitAttr(SrqAttr())
        SRQ(self.pd_res.parent_domain, srq_init_attr)
        cq_init_attrs_ex = CqInitAttrEx(comp_mask=e.IBV_CQ_INIT_ATTR_MASK_PD,
                                        parent_domain=self.pd_res.parent_domain)
        CQEX(self.pd_res.ctx, cq_init_attrs_ex)

    def test_without_allocators(self):
        self._create_parent_domain_with_allocators(None, None)
        self._create_rdma_objects()
        self.pd_res.parent_domain.close()

    def test_default_allocators(self):
        def alloc_p_func(pd, context, size, alignment, resource_type):
            return e._IBV_ALLOCATOR_USE_DEFAULT

        def free_p_func(pd, context, ptr, resource_type):
            return e._IBV_ALLOCATOR_USE_DEFAULT

        self._create_parent_domain_with_allocators(alloc_p_func, free_p_func)
        self._create_rdma_objects()
        self.pd_res.parent_domain.close()

    def test_mem_align_allocators(self):
        def alloc_p_func(pd, context, size, alignment, resource_type):
            p = mem.posix_memalign(size, alignment)
            return p

        def free_p_func(pd, context, ptr, resource_type):
            mem.free(ptr)

        self._create_parent_domain_with_allocators(alloc_p_func, free_p_func)
        self._create_rdma_objects()
        self.pd_res.parent_domain.close()
